#!/usr/bin/env python

import sys
import os
import numpy as np
import copy
import string
from math import sqrt, sin
# import Bio.PDB as bio

from oxDNA_analysis_tools.UTILS.pdb import Atom, Nucleotide, AminoAcid, FROM_OXDNA_TO_ANGSTROM

import oxDNA_analysis_tools.UTILS.protein_to_pdb as pro
import oxDNA_analysis_tools.UTILS.base as base
import oxDNA_analysis_tools.UTILS.utils as utils

from oxDNA_analysis_tools.UTILS.readers import LorenzoReader2

DD12_PDB_PATH = "./UTILS/dd12_na.pdb"
protein_pdb_files = []
def print_usage():
        print("USAGE:", file=sys.stderr)
        print("\t%s topology configuration direction pdbfiles(optional, needs flag -p) " % sys.argv[0], file =sys.stderr)
        print("\t[-H\--hydrogens=True] [-u\--uniform-residue-names] [-p\--contains-protein]", file=sys.stderr)
        exit(1)

def parse_options():
    shortArgs = 'H:uosp'
    longArgs = ['hydrogens=','uniform-residue-names', 'one-file-per-strand', 'contains-protein']
    
    opts = {
        "configuration" : "",
        "topology" : "",
        "oxDNA_direction" : True,
        "print_hydrogens" : True,
        "uniform_residue_names" : False,
        "one_file_per_strand" : False,
        "contains_protein" : False,
    }
    
    try:
        import getopt
        args, positional_args = getopt.gnu_getopt(sys.argv[1:], shortArgs, longArgs)
        for k in args:
            if k[0] == '-H' or k[0] == '--hydrogens':
                k_arg = k[1].lower()
                if k_arg.lower() == "true":
                    opts["print_hydrogens"] = True
                elif k_arg == "false":
                    print >> sys.stderr, "## Hydrogen atoms will *not* be printed"
                    opts["print_hydrogens"] = False
                else:
                    print >> sys.stderr, "The argument of '%s' should be either 'true' or 'false' (got '%s' instead)" % (k[0], k[1])
                    exit(1)
            elif k[0] == '-u' or k[0] == '--uniform-residue-names':
                    opts["uniform_residue_names"] = True
            elif k[0] == '-o' or k[0] == '--one-file-per-strand':
                    opts["one_file_per_strand"] = True
            elif k[0] == '-s' or k[0] == '--same-pdb-all-protein-strands':
                    opts["same_pdb_all_protein_strands"] = True
            elif k[0] == '-p' or k[0] == '--contains-protein':
                    opts["contains_protein"] = True
            
        opts['topology'] = positional_args[0]
        opts['configuration'] = positional_args[1]
        direction = positional_args[2]
        if opts["contains_protein"]:
            protein_pdb_files = positional_args[3:]
        else:
            protein_pdb_files=[]
        
        if direction == "35":
            opts["oxDNA_direction"] = True
        elif direction == "53":
            opts["oxDNA_direction"] = False
        else:
            print >> sys.stderr, "The 'direction' argument should be either 35 or 53"
            exit(1)
    except Exception:
        print_usage()
        
    return opts, protein_pdb_files

def align(full_base, ox_base):
        theta = utils.get_angle(full_base.a3, ox_base._a3)
        # if the two bases are already essentially aligned then we do nothing
        if sin(theta) > 1e-3:
            axis = np.cross(full_base.a3, ox_base._a3)
            axis /= sqrt(np.dot(axis, axis))
            R = utils.get_rotation_matrix(axis, theta)
            full_base.rotate(R)
    
        theta = utils.get_angle(full_base.a1, ox_base._a1)
        if sin(theta) > 1e-3:
            axis = np.cross(full_base.a1, ox_base._a1)
            axis /= sqrt(np.dot(axis, axis))
            R = utils.get_rotation_matrix(axis, theta)
            full_base.rotate(R)

def main():
    opts, protein_pdb_files = parse_options()
        
    # Open PDB File of nice lookin duplex
    with open(os.path.join(os.path.dirname(__file__), DD12_PDB_PATH)) as f:
        nucleotides = []
        aminoacids = []
        old_residue = ""
        for line in f.readlines():
            if len(line) > 77:
                na = Atom(line)
                if na.residue_idx != old_residue:
                    nn = Nucleotide(na.residue, na.residue_idx)
                    nucleotides.append(nn)
                    old_residue = na.residue_idx
                nn.add_atom(na)

    bases = {}
    for n in nucleotides:
        n.compute_as()
        if n.base in bases:
            if n.check < bases[n.base].check: bases[n.base] = copy.deepcopy(n)
        else: 
            bases[n.base] = n
    
    for n in nucleotides:
        n.a1, n.a2, n.a3 = utils.get_orthonormalized_base(n.a1, n.a2, n.a3)
    
    #Lorenzo Reader2 Perhaps?
    try:
        lr = LorenzoReader2(opts['configuration'], opts['topology'])
        s = lr._get_system()
    except Exception as e:
        print("Parser error: %s" % e, file=sys.stderr)
        exit(1)
    
    ox_nucleotides = []
    s.map_nucleotides_to_strands()
    com = np.array([0., 0., 0.])
    for my_strand in s._strands:
        com += my_strand.cm_pos
    com /= s.N_strands
    
    box_angstrom = s._box * FROM_OXDNA_TO_ANGSTROM
    correct_for_large_boxes = False
    if np.any(box_angstrom[box_angstrom > 999]):
        print("At least one of the box sizes is larger than 999: all the atoms which are outside of the box will be brought back through periodic boundary conditions", file=sys.stderr)
        correct_for_large_boxes = True
    
    if opts['one_file_per_strand']:
        out_name = opts['configuration'] + "_1.pdb"
    else:
        out_name = opts['configuration'] + ".pdb"
        
    out = open(out_name, "w+")
    
    current_base_identifier = 'A'   
    reading_position = 0 
    s_pdbfile = iter(protein_pdb_files)
    pdbfile = next(s_pdbfile)
    for s_id, strand in enumerate(s._strands):
        strand_pdb = []
        nucleotides_in_strand = strand._nucleotides
        if not opts['oxDNA_direction']:
            nucleotides_in_strand = reversed(nucleotides_in_strand)
        #Protein
        sys.stdout.write("\rINFO: Converting Strand %i" % strand.index)
        sys.stdout.flush()
        if strand.index < 0:
            #s_pdbfile = protein_pdb_files[abs(strand.index)-1]
            coord = [n.cm_pos for n in strand._nucleotides]  # amino acids only go from nterm to cterm (pdb format does as well)
            #print('oxpositions ', coord)
            #print('reading_pos', reading_position)
            next_reading_position = pro.oxdna_to_pdb(out, coord, pdbfile, com, reading_position)
            if next_reading_position == -1:
                try:
                    pdbfile = next(s_pdbfile)
                    reading_position = 0
                except StopIteration:
                    continue
            #if pdbfile == "end":
                #break
            #else:
                #reading_position = 0
            else:
                 reading_position = next_reading_position

        # DNA
        elif strand.index >= 0:
            for n_idx, nucleotide in enumerate(nucleotides_in_strand, 1):
                nb = base.number_to_base[nucleotide._base]
                my_base = copy.deepcopy(bases[nb])
                my_base.chain_id = s._nucleotide_to_strand[nucleotide.index]
                residue_type = ""
                
                # 3' end
                if nucleotide == strand._nucleotides[0] and not strand._circular:
                    residue_type = "3"
                # 5' end
                elif nucleotide == strand._nucleotides[-1]:
                    residue_type = "5" 
                    
                if opts["uniform_residue_names"] == True:
                    residue_suffix = ""
                else:
                    residue_suffix = residue_type

                align(my_base, nucleotide)
                my_base.set_base((nucleotide.pos_base - com) * FROM_OXDNA_TO_ANGSTROM)

                if correct_for_large_boxes:
                    my_base.correct_for_large_boxes(box_angstrom)
                
                residue_serial = n_idx % 9999
                base_identifier = current_base_identifier
                # Make nucleotide line from pdb.py
                nucleotide_pdb = my_base.to_pdb(base_identifier, opts['print_hydrogens'], residue_serial, residue_suffix, residue_type)
                # Append to strand_pdb
                strand_pdb.append(nucleotide_pdb)
            
            print("\n".join(x for x in strand_pdb), file=out)
            print("TER", file=out)
        
        if opts['one_file_per_strand']:
            out.close()
            print >> sys.stderr, "## Wrote strand %d's data to '%s'" % (s_id + 1, out_name)
            # open a new file if needed
            if strand != s._strands[-1]:
                out_name = opts['configuration'] + "_%d.pdb" % (s_id + 2, )
                out = open(out_name, "w")
        else:
            # we update the base identifier only if a single file is printed
            if current_base_identifier == 'Z':
                current_base_identifier = 'A'
            else:
                current_base_identifier = chr(ord(current_base_identifier) + 1)
    
    # #Must Now renumber and restrand, (sorry)
    out.seek(0)
    outw = open('TM2.pdb', 'w')
    resid, atmid, chainid = -1, 1, -1
    #check against next atom entry
    pres, pchain = 0, 0
    alpha = list(string.ascii_uppercase)
    alpha = list(string.ascii_uppercase)
    a2 = copy.deepcopy(alpha)
    for i in range(26):
        alpha += [a2[i]+x for x in a2]
    for line in out:
        write_chain_end = False
        if line.startswith('ATOM'):
            data = line.split()
            cres_id = data[5]
            curr_chainid = data[4]
            if curr_chainid != pchain:
                if chainid != -1:
                    write_chain_end = True
                chainid += 1
                pchain = curr_chainid 
            if cres_id != pres:
                resid += 1
                pres = cres_id
            data[1] = str(atmid)
            data[5] = str(resid + 1)
            data[4] = alpha[chainid]
        
            coord = line[29:53]
            xd, yd, zd = float(line[30:38]), float(line[38:46]), float(line[46:54])
            for x in [xd, yd, zd]:
                spcs = 7-len(str(x))
                x = ''.join([' ' for i in range(spcs)]) + str(x)
            #print(xd, yd, zd)
            if len(list(data[-1])) == 1:
                del data[-1]

            if len(data) == 8:
                try:
                    bfactor = float(data[7])
                    occ = float(data[6])
                except ValueError:
                    occ = data[7][:4]
                    bfactor = data[7][4:]
            elif len(data) == 9:
                try:
                    bfactor = float(data[8])
                    occ = float(data[7])
                except ValueError:
                    occ = data[8][:4]
                    bfactor = data[8][4:]
            elif len(data) == 10:
                try:
                    bfactor = float(data[9])
                    occ = float(data[8])
                except ValueError:
                    occ = data[9][:4]
                    bfactor = data[9][4:]
            elif len(data) == 11:
                try:
                    bfactor=float(data[10])
                    occ = float(data[9])
                except ValueError:
                    occ = data[10][:4]
                    bfactor = data[10][4:]
            else: 
                bfactor=0
                occ = 1

            for x in [occ, bfactor]:
                spcs = 4-len(str(x))
                x = ''.join([' ' for i in range(spcs)]) + str(x)

            # data indice -> PDB FIELD LENGTH
            dls = {1: 6, 2:4, 3:3, 4:2, 5:6}
            for i in range(1,6):
                x = len(data[i])
                if x != dls[i]:
                    diff = dls[i] - x
                    empty = [' ' for j in range(diff)]
                    data[i] = ''.join(empty)+data[i]
                    #print(data[i])

            
            print('ATOM', data[1], data[2], data[3], data[4], data[5], coord, occ, bfactor, file=outw)
            if write_chain_end:
                print('TER ', data[1], '    ', data[4], data[5], file=outw)
            atmid += 1
            
            
            # nS.init_atom(data[2], [xd, yd, zd], bfactor, occ, alpha[chainid], data[2], serial_number=atmid)
    out.close()
    # io = bio.PDBIO()
    # io.set_structure(nS.get_structure())
    # io.save(out_name)     

    if not opts['one_file_per_strand']:
        outw.close()
        print("\n## Wrote data to '%s'" % out_name, file=sys.stderr)
        
    print("## DONE", file=sys.stderr)

if __name__ == '__main__':
    main()




import numpy as np
import math


try:
    FLT_EPSILON = np.finfo(np.float).eps
except:
    FLT_EPSILON = 2.2204460492503131e-16


def get_angle(a, b):
    """
    Get angle between a,b

    >>> a = [0, 1, 0]
    >>> b = [0, 0, 1]
    >>> round(get_angle(a,b),3)
    1.571

    """
    ab = np.dot(a, b)
    if ab > (1. - FLT_EPSILON): 
        return 0
    elif ab < (-1. + FLT_EPSILON): 
        return np.pi
    else: 
        return np.arccos(ab)


def get_orthonormalized_base(v1, v2, v3):
    v1_norm2 = np.dot(v1, v1)
    v2_v1 = np.dot(v2, v1)

    v2 -= (v2_v1 / v1_norm2) * v1

    v3_v1 = np.dot(v3, v1)
    v3_v2 = np.dot(v3, v2)
    v2_norm2 = np.dot(v2, v2)

    v3 -= (v3_v1 / v1_norm2) * v1 + (v3_v2 / v2_norm2) * v2

    v1 /= np.sqrt(v1_norm2)
    v2 /= np.sqrt(v2_norm2)
    v3 /= np.sqrt(np.dot(v3, v3))

    return v1, v2, v3


def get_random_vector_in_sphere(r=1):
    r2 = r * r
    v = np.random.uniform(-r, r, 3)

    while np.dot(v, v) > r2:
        v = np.random.uniform(-r, r, 3)

    return v


def get_random_vector():
    ransq = 1.

    while ransq >= 1.:
        ran1 = 1. - 2. * np.random.random()
        ran2 = 1. - 2. * np.random.random()
        ransq = ran1 * ran1 + ran2 * ran2

    ranh = 2. * np.sqrt(1. - ransq)
    return np.array([ran1 * ranh, ran2 * ranh, 1. - 2. * ransq])


def get_random_rotation_matrix():
    v1, v2, v3 = get_orthonormalized_base(get_random_vector(), get_random_vector(), get_random_vector())

    R = np.array([v1, v2, v3])
    # rotations have det == 1
    if np.linalg.det(R) < 0: R = np.array([v2, v1, v3])

    return R


def get_rotation_matrix(axis, angle):
    ct = math.cos(angle)
    st = math.sin(angle)
    olc = 1. - ct
    x, y, z = axis / math.sqrt(np.dot(axis, axis))

    return np.array([[olc * x * x + ct, olc * x * y - st * z, olc * x * z + st * y],
                    [olc * x * y + st * z, olc * y * y + ct, olc * y * z - st * x],
                    [olc * x * z - st * y, olc * y * z + st * x, olc * z * z + ct]])



from copyreg import pickle
from oxDNA_analysis_tools.UTILS.micha_reader import MichaReader
import argparse
from os import getenv, path
from pickle import dumps


def main():
    #handle commandline arguments
    parser = argparse.ArgumentParser(prog = path.basename(__file__), description="Creates an index file of the provided trajectory.")
    parser.add_argument('top', type=str, nargs=1, help="The trajectory file to align")
    parser.add_argument('traj', type=str, nargs=1, help="The trajectory file to align")
    args = parser.parse_args()

    # Parse command line arguments
    top_file = args.top[0]
    traj_file = args.traj[0]

    print("indexing provided trajectory")
    # initiate the reader
    reader = MichaReader(top_file, traj_file)
    print("done")


if __name__ == '__main__':
    main()


#!/usr/bin/env python3
import os
from sys import stderr
from re import finditer

import argparse
import numpy as np

from oxDNA_analysis_tools.external_force_utils.force_reader import write_force_file
from oxDNA_analysis_tools.external_force_utils.forces import mutual_trap

# Copied from ExpertRNA
def parse_dot_bracket(input):
    """
    Converts a dot-bracket string to a list of paired nucleotides

    Parameters:
        input (str): A dot-bracket string

    Returns:
        output (list): A list where each index corresponds to a nucleotide.  Value is -1 is unpaired or another index if paired.
    """
    output = np.full(len(input), -1)
    #I'm not sure this is the most efficent way to do this, but I'm lazy.
    more = True
    while more:
        more = False

        #finds matched parenthesis
        for x in finditer(r"\([^()]*\)", input):
            more = True
            output[x.start()] = x.end()-1
            output[x.end()-1] = x.start()

            #its recursive...
            input=input[0:x.start()] + "." + input[x.start()+1:x.end()-1] + "." + input[x.end():]

    return output


def main():
    parser = argparse.ArgumentParser(prog = os.path.basename(__file__), description="Create an external forces file enforcing the current base-pairing arrangement")
    parser.add_argument('db_file', type=str, nargs=1, help="A text file containing dot-bracket notation of the base-pairing arrangement")
    parser.add_argument('-o', '--output', type=str, nargs=1, help='Name of the file to write the force list to')
    parser.add_argument('-s', '--strength', type=float, nargs=1, help='Strength of the forces')
    args = parser.parse_args()

    from oxDNA_analysis_tools.config import check_dependencies
    check_dependencies(["python", "numpy"])

    # Get input
    with open(args.db_file[0]) as f:
        db_str = f.read()

    # Check for strength input, otherwise default to 0.09 which won't explode most simulations.
    if args.strength:
        strength = args.strength[0]
        print("INFO: Using strength {}".format(strength), file=stderr)
    else:
        strength = 0.09
        print("INFO: No strength provided, defaulting to {}".format(strength), file=stderr)

    # convert the db string to an index list
    db_idx = parse_dot_bracket(db_str)

    force_list = []

    #p is particle id, q is paired particle id
    for p, q in enumerate(db_idx):
        if q != -1:
            force_list.append(mutual_trap(p, q, strength, 1.2, 1))

    # write the force file
    if args.output:
        outfile = args.output[0]
        print("INFO: Writing forces to {}".format(outfile), file=stderr)
    else:
        outfile = "external_forces.txt"
        print("INFO: No output filename found.  Defaulting to {}".format(outfile), file=stderr)

    write_force_file(force_list, outfile)

if __name__ == '__main__':
    main()


#see: https://dna.physics.ox.ac.uk/index.php/Documentation#External_Forces


def mutual_trap(particle, ref_particle, stiff, r0, PBC):
    """
    A spring force that pulls a particle towards the position of another particle

    Parameters:
        particle (int): the particle that the force acts upon
        ref_particle (int): the particle that the particle will be pulled towards
        stiff (float): the force constant of the spring (in simulation units)
        r0 (float): the equlibrium distance of the spring
        PBC (bool): does the force calculation take PBC into account (almost always 1)
    """
    PBC = int(PBC)
    return({
        "type" : "mutual_trap",
        "particle" : particle,
        "ref_particle" : ref_particle,
        "stiff" : stiff, 
        "r0" : r0,
        "PBC" : PBC
    })


def string(particle, f0, rate, direction):
    """
    A linear force along a vector

    Parameters:
        particle (int): the particle that the force acts upon
        f0 (float): the initial strength of the force at t=0 (in simulation units)
        rate (float or SN string): growing rate of the force (simulation units/timestep)
        dir ([float, float, float]): the direction of the force
    """
    return({
        "type" : "string",
        "particle" : particle, 
        "f0" : f0, 
        "rate" : rate, 
        "dir" : direction 
    })


def harmonic_trap(particle, pos0, stiff, rate, direction):
    """
    A linear potential well that traps a particle

    Parameters:
        particle (int): the particle that the force acts upon
        pos0 ([float, float, float]): the position of the trap at t=0
        stiff (float): the stiffness of the trap (force = stiff * dx)
        rate (float): the velocity of the trap (simulation units/time step)
        direction ([float, float, float]): the direction of movement of the trap
    """
    return({
        "type" : "trap",
        "particle" : particle, 
        "pos0" : pos0,
        "rate" : rate,
        "dir" : direction
    })


def rotating_harmonic_trap(particle, stiff, rate, base, pos0, center, axis, mask):
    """
    A harmonic trap that rotates in space with constant angular velocity

    Parameters:
        particle (int): the particle that the force acts upon
        pos0 ([float, float, float]): the position of the trap at t=0
        stiff (float): the stiffness of the trap (force = stiff * dx)
        rate (float): the angular velocity of the trap (simulation units/time step)
        base (float): initial phase of the trap
        axis ([float, float, float]): the rotation axis of the trap
        mask([float, float, float]): the masking vector of the trap (force vector is element-wise multiplied by mask)
    """
    return({
        "type" : "twist", 
        "particle" : particle,
        "stiff" : stiff,
        "rate" : rate,
        "base" : base,
        "pos0" : pos0,
        "center" : center,
        "axis" : axis,
        "mask" : mask
    })


def repulsion_plane(particle, stiff, direction, position):
    """
    A plane that forces the affected particle to stay on one side.

    Parameters:
        particle (int): the particle that the force acts upon.  -1 will act on whole system.
        stiff (float): the stiffness of the trap (force = stiff * distance below plane)
        dir ([float, float, float]): the normal vecor to the plane
        position(float): position of the plane (plane is d0*x + d1*y + d2*z + position = 0)
    """
    return({
        "type" : "repulsion_plane",
        "particle" : particle,
        "stiff" : stiff,
        "dir" : direction,
        "position" : position
    })


def repulsion_sphere(particle, center, stiff, r0, rate=1):
    """
    A sphere that encloses the particle
    
    Parameters:
        particle (int): the particle that the force acts upon
        center ([float, float, float]): the center of the sphere
        stiff (float): stiffness of trap
        r0 (float): radius of sphere at t=0
        rate (float): the sphere's radius changes to r = r0 + rate*t
    """
    return({
        "type" : "sphere",
        "center" : center,
        "stiff" : stiff,
        "r0" : r0,
        "rate" : rate
    })



from setuptools import setup, find_packages
import re
import os

def get_property(prop):
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), open('src/oxDNA_analysis_tools/__init__.py').read())
    return result.group(1)

setup(version = get_property('__version__'))

print("\n\n################################################################################")
print("\nAll set up! All scripts in the package can now be run from the command line with\n\n    oat <script_name>\n\nimporting functions in your own scripts can be done with\n\n    from oxDNA_analysis_tools.<file name> import <function>")

print("")

print("If you would like to enable autocomplete in Bash, either add\n\n    source oat-completion.sh\n\nto your .bashrc or run \n\n    cat oat-completion.sh >> ~/.bash_completion\n\nto make autocompletes available to all users.")



import numpy as np
import itertools
from math import sqrt
import sys
import copy

BASE_SHIFT = 1.13
COM_SHIFT = 0.5
FROM_OXDNA_TO_ANGSTROM = 8.518
FROM_ANGSTROM_TO_OXDNA = 1. / FROM_OXDNA_TO_ANGSTROM

NAME_TO_BASE = {
        "ADE" : "A",
        "CYT" : "C",
        "GUA" : "G",
        "THY" : "T",
        "URA" : "U",
    }
NAME_TO_AA = {'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K','ILE': 'I', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ASN': 'N', 
                'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W', 'ALA': 'A', 'VAL':'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M'}
aa_to_number = {'A':-1, 'R':-2, 'N':-3, 'D':-4, 'C':-5, 'E':-6, 'Q':-7, 'G':-8, 'H':-9, 'I':-10, 'L':-11, 'K':-12, 'M':-13, 'F':-14, 'P':-15, 'S':-16, 'T':-17, 'W':-18, 'Y':-19, 'V':-20, 'Z':-21, 'X':0}
BASES = ["A", "T", "G", "C", "U"]
AAS = ["A", "R", "N", "D", "C", "E", "Q", "G", "H", "I", "L", "K", "M", "F", "P", "S", "T", "W", "Y", "V"]

class Nucleotide(object):
    RNA_warning_printed = False
    
    def __init__(self, name, idx):
        object.__init__(self)
        self.name = name.strip()
        if self.name in NAME_TO_BASE.keys():
            self.base = NAME_TO_BASE[self.name]
        elif self.name in BASES:
            if self.name == "U" and not Nucleotide.RNA_warning_printed:
                print >> sys.stderr, "WARNING: unsupported uracil detected: use at your own risk"
                Nucleotide.RNA_warning_printed = True
                
            self.base = self.name
        else:
            self.base = name[1:]
        self.idx = idx
        self.base_atoms = []
        self.phosphate_atoms = []
        self.sugar_atoms = []
        self.named_atoms = {}
        self.ring_names = ["C2", "C4", "C5", "C6", "N1", "N3"]
        self.chain_id = None

    def get_atoms(self):
        return self.base_atoms + self.phosphate_atoms + self.sugar_atoms

    def add_atom(self, a):
        if 'P' in a.name or a.name == "HO5'": 
            self.phosphate_atoms.append(a)
        elif "'" in a.name: 
            self.sugar_atoms.append(a)
        else: 
            self.base_atoms.append(a)
        
        self.named_atoms[a.name] = a
        if self.chain_id == None: 
            self.chain_id = a.chain_id

    def get_com(self, atoms=None):
        if atoms == None: 
            atoms = self.atoms
        com = np.array([0., 0., 0.])
        for a in atoms:
            com += a.pos

        return com / len(atoms)

    def compute_a3(self):
        base_com = self.get_com(self.base_atoms)
        # the O4' oxygen is always (at least for non pathological configurations, as far as I know) oriented 3' -> 5' with respect to the base's centre of mass
        parallel_to = self.named_atoms["O4'"].pos - base_com
        self.a3 = np.array([0., 0., 0.])
        
        for perm in itertools.permutations(self.ring_names, 3):
            p = self.named_atoms[perm[0]]
            q = self.named_atoms[perm[1]]
            r = self.named_atoms[perm[2]]
            v1 = p.pos - q.pos
            v2 = p.pos - r.pos
            v1 /= sqrt(np.dot(v1, v1))
            v2 /= sqrt(np.dot(v2, v2))
            if abs(np.dot(v1, v2)) > 0.01 or 1:
                a3 = np.cross(v1, v2)
                a3 /= sqrt(np.dot(a3, a3))
                if np.dot(a3, parallel_to) < 0.: 
                    a3 = -a3
                self.a3 += a3

        self.a3 /= sqrt(np.dot(self.a3, self.a3))

    def compute_a1(self):
        if "C" in self.name or "T" in self.name or "U" in self.name:
            pairs = [ ["N3", "C6"], ["C2", "N1"], ["C4", "C5"] ]
        else:
            pairs = [ ["N1", "C4"], ["C2", "N3"], ["C6", "C5"] ]

        self.a1 = np.array([0., 0., 0.])
        for pair in pairs:
            p = self.named_atoms[pair[0]]
            q = self.named_atoms[pair[1]]
            diff = p.pos - q.pos
            self.a1 += diff

        self.a1 /= sqrt(np.dot(self.a1, self.a1))

    def compute_as(self):
        self.compute_a1()
        self.compute_a3()
        self.a2 = np.cross(self.a3, self.a1)
        self.check = abs(np.dot(self.a1, self.a3))
        
    def correct_for_large_boxes(self, box):
        map(lambda x: x.shift(-np.rint(x.pos / box ) * box), self.atoms)

    def to_pdb(self, chain_identifier, print_H, residue_serial, residue_suffix, residue_type):
        res = []
        for a in self.atoms:
            if not print_H and 'H' in a.name:
                continue
            if residue_type == "5": 
                if 'P' in a.name:
                    if a.name == 'P':
                        phosphorus = a
                    continue
                elif a.name == "O5'":
                    O5prime = a
            elif residue_type == "3":
                if a.name == "O3'":
                    O3prime = a
            res.append(a.to_pdb(chain_identifier, residue_serial, residue_suffix))
            
        # if the residue is a 3' or 5' end, it requires one more hydrogen linked to the O3' or O5', respectively
        if residue_type == "5":
            new_hydrogen = copy.deepcopy(phosphorus)
            new_hydrogen.name = "HO5'"
            
            # we put the new hydrogen at a distance 1 Angstrom from the O5' oxygen along the direction that, in a regular nucleotide, connects O5' and P
            dist_P_O = phosphorus.pos - O5prime.pos
            dist_P_O *= 1. / np.sqrt(np.dot(dist_P_O, dist_P_O))
            new_hydrogen.pos = O5prime.pos + dist_P_O
            res.append(new_hydrogen.to_pdb(chain_identifier, residue_serial, residue_suffix))
        elif residue_type == "3":
            new_hydrogen = copy.deepcopy(O3prime)
            new_hydrogen.name = "HO3'"
            
            # we put the new hydrogen at a distance 1 Angstrom from the O3' oxygen along a direction which is a linear combination of the three 
            # orientations that approximately reproduce the crystallographic position
            new_distance = 0.2 * self.a2 - 0.2 * self.a1 - self.a3
            new_distance *= 1. / np.sqrt(np.dot(new_distance, new_distance))
            new_hydrogen.pos = O3prime.pos + new_distance
            res.append(new_hydrogen.to_pdb(chain_identifier, residue_serial, residue_suffix))

        return "\n".join(res)

    def to_mgl(self):
        res = []
        for a in self.atoms:
            res.append(a.to_mgl())

        return "\n".join(res)

    def rotate(self, R):
        com = self.get_com()
        for a in self.atoms:
            a.pos = np.dot(R, a.pos - com) + com

        self.compute_as()

    def set_com(self, new_com):
        com = self.get_com()
        for a in self.atoms:
            a.pos += new_com - com - COM_SHIFT * self.a1

    def set_base(self, new_base_com):
        atoms = [v for k, v in self.named_atoms.items() if k in self.ring_names]
        ring_com = self.get_com(atoms)
        for a in self.atoms:
            a.pos += new_base_com - ring_com - BASE_SHIFT * self.a1

        self.compute_as()

    atoms = property(get_atoms)


class Atom(object):
    serial_atom = 1

    def __init__(self, pdb_line):
        object.__init__(self)
        # http://cupnet.net/pdb-format/
        self.name = pdb_line[12:16].strip()
        # some PDB files have * in place of '
        if "*" in self.name:
            self.name = self.name.replace("*", "'")
        
        self.alternate = pdb_line[16]
        self.residue = pdb_line[17:20].strip()
        self.chain_id = pdb_line[21:22].strip()
        self.residue_idx = int(pdb_line[22:26])
        self.pos = np.array([float(pdb_line[30:38]), float(pdb_line[38:46]), float(pdb_line[46:54])])
        
    def is_hydrogen(self):
        return "H" in self.name

    def shift(self, diff):
        self.pos += diff

    def to_pdb(self, chain_identifier, residue_serial, residue_suffix):
        residue = self.residue + residue_suffix
        res = "{:6s}{:5d} {:^4s}{:1s}{:3s} {:1s}{:1s}{:4d}{:1s}  {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}          {:>2s}{:2s}".format("ATOM", Atom.serial_atom, self.name, " ", residue, chain_identifier," ", residue_serial, " ", self.pos[0], self.pos[1], self.pos[2], 1.00, 0.00, " ", " ", " ")
        Atom.serial_atom += 1
        if Atom.serial_atom > 99999:
            Atom.serial_atom = 1
        return res

    def to_mgl(self):
        colors = {"C" : "0,1,1", "P" : "1,1,0", "O" : "1,0,0", "H" : "0.5,0.5,0.5", "N" : "0,0,1"}
        for c in colors:
            if c in self.name: color = colors[c]
        r = 0.5
        return "%s %s %s @ %f C[%s]" % (self.pos[0], self.pos[1], self.pos[2], r, color)


class AminoAcid(object):
    # RNA_warning_printed = False
    
    def __init__(self, name, idx):
        object.__init__(self)
        self.name = name.strip()
        if self.name in NAME_TO_AA.keys():
            self.base = NAME_TO_AA[self.name]
        elif self.name in AAS:
            self.base = self.name
        else:
            self.base = name[1:]
        self.idx = idx
        self.backbone_atoms = []
        self.sidechain_atoms = []
        self.ca_atom = None
        self.chain_id = None

    def get_atoms(self):
        return self.backbone_atoms + self.sidechain_atoms + self.ca_atom

    def add_atom(self, a):
        if a.name in ['N', 'C', 'O']:
            self.backbone_atoms.append(a)
        elif a.name == 'CA': 
            self.ca_atom = a
        else: 
            self.sidechain_atoms.append(a)
        
        if self.chain_id == None: 
            self.chain_id = a.chain_id

    def get_ca_pos(self):
        # Use Residue's Alpha Carbon position
        if self.ca_atom:
            com = np.array(self.ca_atom.pos)

        return com

    def get_com(self, atoms=None):
        if atoms == None: 
            self.atoms = self.get_atoms()
            atoms = self.atoms
        com = np.array([0., 0., 0.])
        for a in atoms:
            com += a.pos

        return com / len(atoms)

    # def compute_as(self):
        # self.compute_a1()
        # self.compute_a3()
        # self.a2 = np.cross(self.a3, self.a1)
        # self.check = abs(np.dot(self.a1, self.a3))
        
    def correct_for_large_boxes(self, box):
        map(lambda x: x.shift(-np.rint(x.pos / box ) * box), self.atoms)

    def to_pdb(self, chain_identifier, print_H, residue_serial, residue_suffix, residue_type):
        res = []
        for a in self.atoms:
            res.append(a.to_pdb(chain_identifier, residue_serial, residue_suffix))
            
        # if the residue is a 3' or 5' end, it requires one more hydrogen linked to the O3' or O5', respectively
        # if residue_type == "5":
        #     new_hydrogen = copy.deepcopy(phosphorus)
        #     new_hydrogen.name = "HO5'"
            
        #     # we put the new hydrogen at a distance 1 Angstrom from the O5' oxygen along the direction that, in a regular nucleotide, connects O5' and P
        #     dist_P_O = phosphorus.pos - O5prime.pos
        #     dist_P_O *= 1. / np.sqrt(np.dot(dist_P_O, dist_P_O))
        #     new_hydrogen.pos = O5prime.pos + dist_P_O
        #     res.append(new_hydrogen.to_pdb(chain_identifier, residue_serial, residue_suffix))
        # elif residue_type == "3":
        #     new_hydrogen = copy.deepcopy(O3prime)
        #     new_hydrogen.name = "HO3'"
            
        #     # we put the new hydrogen at a distance 1 Angstrom from the O3' oxygen along a direction which is a linear combination of the three 
        #     # orientations that approximately reproduce the crystallographic position
        #     new_distance = 0.2 * self.a2 - 0.2 * self.a1 - self.a3
        #     new_distance *= 1. / np.sqrt(np.dot(new_distance, new_distance))
        #     new_hydrogen.pos = O3prime.pos + new_distance
        #     res.append(new_hydrogen.to_pdb(chain_identifier, residue_serial, residue_suffix))

        return "\n".join(res)

    # def to_mgl(self):
    #     res = []
    #     for a in self.atoms:
    #         res.append(a.to_mgl())

    #     return "\n".join(res)

    def rotate(self, R):
        com = self.get_com()
        for a in self.atoms:
            a.pos = np.dot(R, a.pos - com) + com


    def set_com(self, new_com):
        com = self.get_com()
        for a in self.atoms:
            a.pos += new_com - com - COM_SHIFT * self.a1

    # def set_base(self, new_base_com):
    #     atoms = [v for k, v in self.named_atoms.iteritems() if k in self.ring_names]
    #     ring_com = self.get_com(atoms)
    #     for a in self.atoms:
    #         a.pos += new_base_com - ring_com - BASE_SHIFT * self.a1

    #     self.compute_as()

    atoms = property(get_atoms)


import Bio
import Bio.PDB
import numpy as np
import numpy.linalg as la
import copy


conv={'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K','ILE': 'I', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ASN': 'N',
                'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W', 'ALA': 'A', 'VAL':'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M'}

                
def getalphacarbons_pdb(pdbid,pdbfile):
        structure = Bio.PDB.PDBParser().get_structure(pdbid, pdbfile)
        model = structure[0]
        # #chain index, residue index, residue identitiy, CA coordinates
        # cindx, rindx, rids, coord = [], [], [], []
        # if len(model) ==1:
        #     self.single_chain=True
        # else:
        #     self.single_chain=False
        calphacoord, ocoord, c_rids, c_atomids = [], [], [], []
        for cid, chain in enumerate(model):
            rids = []
            atomids = []
            res_indx =0
            for residue in chain.get_residues():
                rids.append(res_indx)
                tags= residue.get_full_id()
                if tags[3][0] == " ":
                    atoms=residue.get_atoms()
                    # Check for alpha carbon in residue
                    ac = [x for x in atoms if x.get_id() == 'CA']
                    if ac:
                        #Shouldn't have to recall this but won't work unless i do
                        atoms=residue.get_atoms()
                        calpha = ac[0].get_coord()
                        calphacoord.append(calpha)
                        for atom in atoms:
                            atomids.append(res_indx)
                            if atom.get_id() == 'CA':
                                ocoord.append(calpha)
                            else:
                                other = atom.get_coord()
                                ocoord.append(other) 
                        res_indx += 1   
            c_rids.append(rids)
            c_atomids.append(atomids) 
        npca = np.asarray(calphacoord)
        npoa = np.asarray(ocoord)
        #print(npca.shape, npoa.shape, len(rids))
        return npca, npoa, c_rids, c_atomids

def get_positions_oxfiles(datfile, conv=True):
    m = open(datfile, 'r')
    datdata = m.readlines()[3:]
    N = len(datdata)
    m.close()
    pos = []
    for line in datdata:
        if conv:
            ptmp = [8.518 * float(x) for x in line.split()[0:3] ]
        else:
            ptmp = [ float(x) for x in line.split()[0:2] ]
        pos.append(ptmp)
    nppos = np.asarray(pos)
    return nppos, N

# All units are Angstroms
# oca, N = get_positions_oxfiles('rotated.dat')
# pca, pfull, residue_ids = getalphacarbons_pdb('sgfp', 'sgfp.pdb')

def get_centroid(npcoords):
    centroid = np.mean(npcoords, axis=0)
    return centroid

# cs = get_centroid(oca)
# ct = get_centroid(pca)

def offset_coord(npcoords, centroid):
    off = np.subtract(npcoords, centroid)
    return off

#Adjust Coordinates to center at (0,0,0)
# oadj = offset_coord(oca, cs)
# padj = offset_coord(pca, ct)

# fulladj = offset_coord(pfull, ct)

# # Cross Covariance Matrix
# M = np.dot(padj.T,oadj)
# # Singular Value Decomposition
# u, s, v = la.svd(M, compute_uv=True)
# # Get Rotation Matrix
# R = np.dot(v.T, u.T)

def update_fullpdb_positions(R, fulladj, pcent, ocent):
    N = fulladj.shape[0]
    fullfit = np.full((N, 3), 0.0)
    for i in range(N):
        fullfit[i] = np.add(ocent, np.dot(R, fulladj[i]))
    return fullfit

def update_core_positions(R, padj, pcent, ocent):
    N = padj.shape[0]
    corefit = np.full((N, 3), 0.0)
    for i in range(N):
        corefit[i] = np.dot(R, padj[i])
    return corefit

# fullp = update_fullpdb_positions(R, fulladj, ct, cs)
# corep = update_core_positions(R, padj, ct, cs)

def slide_pdb_to_oxcoordinates(corep, oadj, fullp, rids):
    # rids is the residue each atom belongs to
    #corepositions are the unshifted calpha coordinates with centroid at (0,0,0)
    fulladj = np.full(fullp.shape, 0.0)
    # print(fulladj.shape)
    # print(fullp.shape)
    # print(oadj.shape, corep.shape)
    adj = np.subtract(oadj, corep)  # shift of c alpha position
    # print(adj)
    # print(adj.shape)
    
    for idx, i in enumerate(rids):
        # if i < 10:
        # print(i, idx)
        fulladj[idx] = fullp[idx] + adj[i]
    return fulladj

# print(residue_ids)
# fcoord = slide_pdb_to_oxcoordinates(corep, oadj, fullp, residue_ids)

def prep_pdb_sys(fullcoordinates, original_pdbfile, original_pdbid, old_reading_position, new_reading_position, offset=0, wipe=True):
    structure = Bio.PDB.PDBParser().get_structure(original_pdbid, original_pdbfile)
    model = structure[0]
    chain_total = len(model)
    # print('ctotal', chain_total)
    atm_idx = 0
    removal, chainremoval = [], []
    
    
    chainkeep  = []
    if new_reading_position == -1 and old_reading_position == 0:
        chainkeep += [x for x in range(chain_total)]  # entire file
    elif new_reading_position - old_reading_position == 1 or new_reading_position == -1: #just a specific chain
        chainkeep.append(old_reading_position)
    
    # print('chainkeep', chainkeep)
    for cid, chain in enumerate(model):
        # print(chain.get_id())
        if cid not in chainkeep:
            chainremoval.append(chain.get_id())
        else:
            for residue in chain.get_residues():
                tags = residue.get_full_id()
                if tags[3][0] == " ":
                    atoms = residue.get_atoms()
                    ac = [x for x in atoms if x.get_id() == 'CA']
                    if ac:
                        atoms=residue.get_atoms()
                        for atom in atoms:
                            atom.set_coord(list(fullcoordinates[atm_idx]))
                            atm_idx += 1
                    # elif wipe:
                    #     removal.append(residue.get_id())
                elif wipe:
                    removal.append(residue.get_id())
            if wipe:
                for x in removal:
                    chain.detach_child(x)
    if wipe:
        for x in chainremoval:
            model.detach_child(x)

    return model

def write_pdb(model, filehandle='', pdb_out=''):
    #Save File
    if pdb_out:
        io = Bio.PDB.PDBIO()
        io.set_structure(model)
        io.save(pdb_out+'.pdb')
    elif filehandle:
    # Returns PDB in Text for use in oxdna_to_PDB
        io = Bio.PDB.PDBIO()
        io.set_structure(model)
        io.save(filehandle, write_end=False)


# write_pdb(fcoord, 'sgfp.pdb', 'sgfp', 'sgfp_ox', wipe=True)
# ca, trash, trash2 =  getalphacarbons_pdb('sgfp_ox','sgfp_ox.pdb')
# print(ca[0], oca[0])



#For use in oxDNA_PDB.py Script
#will write when given the filehandle

def oxdna_to_pdb(filehandle, ox_protein_positions, pdbfile, nuccom, reading_position, offset=0):
    if "/" in pdbfile:
        pdbid=pdbfile.rsplit('/',1)[1].split('.')[0]
    else:
        pdbid=pdbfile.split('.')[0]
    # Get Coordinates
    oca = np.multiply(np.asarray(ox_protein_positions), 8.518)  # simulation to Angstroms
    pca, pfull, chain_separated_residue_ids, chain_separated_atom_ids = getalphacarbons_pdb(pdbid, pdbfile)
    
    #print('reading', reading_position)
    tmp = chain_separated_residue_ids[int(reading_position)]
    tmp_atom = chain_separated_atom_ids[int(reading_position)]
    #print('chains', len(chain_separated_residue_ids))
    
    res_ids = []
    new_reading_position = copy.deepcopy(reading_position)

    if reading_position >= len(chain_separated_residue_ids):
        print('trying to read data that doesnt exist')
        return
    elif reading_position == len(chain_separated_residue_ids)-1:
        abs_pos_res = sum([len(x) for x in chain_separated_residue_ids[:reading_position]])
        abs_pos_atom = sum([len(x) for x in chain_separated_atom_ids[:reading_position]])
        res_ids += chain_separated_atom_ids[reading_position]
        # print("ab res", abs_pos_res)
        # print("ab_atom", abs_pos_atom)
        pca = pca[abs_pos_res:abs_pos_res+len(tmp)]
        pfull = pfull[abs_pos_atom:abs_pos_atom+len(tmp_atom)]
        new_reading_position = -1 # done
    elif reading_position == 0:
        # print(len(tmp), len(oca))
        if sum([len(x) for x in chain_separated_residue_ids]) == len(oca):
             pca = pca
             pfull = pfull
             res_ids += [item for chain in chain_separated_atom_ids for item in chain]
             new_reading_position = -1
        elif len(tmp) == len(oca):
             # print("here")
             pca = pca[:len(tmp)]
             pfull = pfull[:len(tmp_atom)]
             res_ids += chain_separated_atom_ids[reading_position]
             new_reading_position += 1
        else:
             print('trying to read data that doesnt exist')
             return
    else:
        abs_pos_res = sum([len(x) for x in chain_separated_residue_ids[:reading_position]])
        abs_pos_atom = sum([len(x) for x in chain_separated_atom_ids[:reading_position]])
        # print("ab res", abs_pos_res)
        # print("ab_atom", abs_pos_atom)
        res_ids += chain_separated_atom_ids[reading_position]
        pca = pca[abs_pos_res:abs_pos_res+len(tmp)]
        pfull = pfull[abs_pos_atom:abs_pos_atom+len(tmp_atom)]
        new_reading_position += 1


    # pfull -> all atoms belonging to chain/model of PDB
    # pca -> CA atoms belonging to chain/model of PDB

    # Get Centroids
    cs = get_centroid(oca)   # COM of CA oxdna coordinates
    ct = get_centroid(pca)   # COM of CA DB coordinates
    
    # Center Centroid (0, 0, 0) in Coordinates for Rotation Matrix Calculation
    oadj = offset_coord(oca, cs)   # centered CA positions of oxDNA coordinates
    padj = offset_coord(pca, ct)   # centered CA positions of PDB coordinates
    
    #All atoms
    #print(pfull)
    fulladj = offset_coord(pfull, ct)
    # print('fulladj', fulladj)

    # Cross Covariance Matrix
    M = np.dot(padj.T,oadj)
    # Singular Value Decomposition
    u, s, v = la.svd(M, compute_uv=True)
    # Get Rotation Matrix
    R = np.dot(v.T, u.T)

    # Rotates and Centers all atoms of PDB to oxDNA Coordinates
    fullp = update_fullpdb_positions(R, fulladj, ct, cs)
    # Rotates and Centers ALphacarbons of PDB to oxDNA Coordinates
    corep = update_core_positions(R, padj, ct, cs)

    #Adjust position from PDB coordinate to oxCoordinate
    fcoord = slide_pdb_to_oxcoordinates(corep, oadj, fullp, res_ids)
    
    #Adjust position to line up with Nucleotides
    fcoord2 = np.subtract(fcoord, 8.518*nuccom)
    
    #Removes Everything from pdbfile but atoms of residues
    sys = prep_pdb_sys(fcoord2, pdbfile, pdbid, reading_position, new_reading_position, offset=offset)

    #write_pdb
    write_pdb(sys, filehandle)
    
    return new_reading_position


